models = {
    "affinitypropagation": "AffinityPropagation",
    "agglomerativeclustering": "AgglomerativeClustering",
    "birch": "Birch",
    "dbscan": "DBSCAN",
    "optics": "OPTICS",
    "kmeans": "KMeans",
    "featureagglomeration": "FeatureAgglomeration",
    "meanshift": "MeanShift",
    "minibatchkmeans": "MiniBatchKMeans",
    "spectralclustering": "SpectralClustering",
    "affinitypropagation": "affinity_propagation",
    "spectralbiclustering": "SpectralBiclustering",
    "spectralcoclustering": "SpectralCoclustering",
}
