models = {
    "binarizer": "Binarizer",
    "functiontransformer": "FunctionTransformer",
    "kbinsdiscretizer": "KBinsDiscretizer",
    "kernelcenterer": "KernelCenterer",
    "labelbinarizer": "LabelBinarizer",
    "labelencoder": "LabelEncoder",
    "multilabelbinarizer": "MultiLabelBinarizer",
    "quantiletransformer": "QuantileTransformer",
    "normalizer": "Normalizer",
    "onehotencoder": "OneHotEncoder",
    "ordinalencoder": "OrdinalEncoder",
    "powertransformer": "PowerTransformer",
    "splinetransformer": "SplineTransformer",
    "minmax": "MinMaxScaler",
    "maxabs": "MaxAbsScaler",
    "robust": "RobustScaler",
    "standard": "StandardScaler",
    "minmaxscaler": "MinMaxScaler",
    "maxabsscaler": "MaxAbsScaler",
    "robustscaler": "RobustScaler",
    "standardscaler": "StandardScaler",
}
